from textual.app import App, ComposeResult
from textual.containers import Container, Vertical, Horizontal
from textual.widgets import Input, Static, Select
from textual.reactive import reactive
from textual.worker import Worker, WorkerState
from nllw.core import TranslationBackend
from nllw.languages import LANGUAGES

class TranslationApp(App):
    """Interactive translation application with Textual."""
    
    CSS = """
    #main-container {
        background: white;
    }

    .lang-row {
        height: auto;
        margin-bottom: 1;
    }

    .column {
        margin-left: 3;
        margin-right: 3
    }

    .lang-label {
        color: $text-muted;
        margin: 0;
    }


    #input-container {
        layout: horizontal;
        border: round $primary;
        padding: 0 1;
        height: 3;
        align: left middle;
    }

    .prompt-symbol {
        color: $primary;
        text-style: bold;
        width: auto;
        height: auto;
    }

    #input-field {
        border: none;
        background: transparent;
        height: 1;
        padding: 0;
        width: 1fr;
    }

    #input-field:focus {
        border: none;
    }

    #output-container {
        border: round $accent;
        padding: 1 2;
        min-height: 8;
    }

    #output {
        padding: 0;
        min-height: 3;
    }
    """
    
    BINDINGS = [
        ("q", "quit", "Quit"),
        ("ctrl+c", "quit", "Quit"),
        ("ctrl+t", "toggle_theme", "Toggle Theme"),
    ]
    
    output_text = reactive("")
    
    def __init__(self):
        super().__init__()
        self.backend = None
        self.backend_loading = False
        self.last_words_count = 0
        self.theme = "catppuccin-latte"
        self.current_worker = None
        self.debug_log = []
        self.current_input_text = ""

        self.source_lang = "fra_Latn"
        self.target_lang = "eng_Latn"
            
    def compose(self) -> ComposeResult:
        with Container(id="main-container"):
            yield Horizontal(
                Vertical(
                    Static("From:", classes="lang-label"),
                    Select(
                        ((language['name'], language['nllb']) for language in LANGUAGES),
                        value="fra_Latn",
                        id="source-lang", 
                        compact=True
                    ),
                    classes="column",
                ),
                Vertical(
                    Static("To:", classes="lang-label"),
                    Select(
                        ((language['name'], language['nllb']) for language in LANGUAGES),
                        value="eng_Latn",
                        id="target-lang",
                        compact=True
                    ),
                    classes="column",
                ),
            )

            with Container(id="input-container"):
                yield Static("> ", classes="prompt-symbol")
                yield Input(
                    placeholder="Type your text here...",
                    id="input-field"
                )

            with Container(id="output-container"):
                yield Static(id="output")
    
    def on_mount(self) -> None:
        self._load_backend()
    
    def action_toggle_theme(self) -> None:
        self.theme = "catppuccin-latte" if self.theme == "textual-dark" else "textual-dark"
    
    def _load_backend(self) -> None:
        if self.backend is None and not self.backend_loading:
            self.backend_loading = True
            try:
                self.query_one("#output", Static).update(
                    "[yellow]Loading translation model...[/]"
                )
                self.backend = TranslationBackend(
                    source_lang=self.source_lang,
                    target_lang=self.target_lang
                )
                self.query_one("#output", Static).update(
                    f"[green]Type {self.source_lang} text and press space to translate to {self.target_lang}.[/]"
                )
            except Exception as e:
                self.query_one("#output", Static).update(
                    f"[red]Error loading model: {str(e)}[/]"
                )
                self.backend = None
            finally:
                self.backend_loading = False
    
    def on_input_changed(self, event: Input.Changed) -> None:
        input_text = event.value
        
        if self.current_worker is not None and self.current_worker.state == WorkerState.RUNNING:
            self.current_worker.cancel()
        
        if not self._should_translate(input_text):
            status_text = self._get_status_text(input_text)
            if status_text:
                self.query_one("#output", Static).update(status_text)
        else:
            self.current_input_text = input_text
            self.current_worker = self.run_worker(
                lambda: self._translate_async(input_text),
                thread=True,
                exclusive=True
            )
    
    def _should_translate(self, text: str) -> bool:
        if not text or self.backend_loading or self.backend is None:
            return False
        
        word_count = len(text.strip().split())
        if word_count < 3:
            return False
        
        if text.endswith(' '):
            if word_count >= self.last_words_count + 1:
                self.last_words_count = word_count
                return True
        
        return False
    
    def _get_status_text(self, text: str) -> str:
        if not text:
            return "[dim italic]Waiting for text...[/]"
        
        if self.backend_loading:
            return "[yellow]Loading model...[/]"
        
        word_count = len(text.strip().split())
        if word_count < 3:
            return f"[dim italic]Type at least 3 words to start translation... ({word_count}/3)[/]"
        
        return None
    
    def _translate_async(self, text: str) -> tuple:
        stable_translation, buffer = self.backend.translate(text)
        return stable_translation, buffer, text
    
    def on_select_changed(self, event: Select.Changed) -> None:
        """Handle language selection changes."""
        if event.select.id == "source-lang":
            self.source_lang = str(event.value)
        elif event.select.id == "target-lang":
            self.target_lang = str(event.value)        
        if hasattr(self, 'backend') and self.backend is not None:
            self.backend = None
            # self.debug_log.clear()
            # self.query_one("#debug-output", Static).update("")
            self._load_backend()
    
    def on_worker_state_changed(self, event: Worker.StateChanged) -> None:
        """Handle worker state changes and update UI with results."""
        if event.state == WorkerState.SUCCESS:
            stable_translation, buffer, original_text = event.worker.result            
            
            if not stable_translation and not buffer:
                return
            
            output = stable_translation
            if buffer:
                output += f"[$accent]{buffer}[/]"
            
            self.query_one("#output", Static).update(output)
            
            # debug_entry = f"{original_text}: {stable_translation} | {buffer}"
            # self.debug_log.append(debug_entry)
            # debug_text = "\n".join(self.debug_log)
            # self.query_one("#debug-output", Static).update(debug_text)
        elif event.state == WorkerState.ERROR:
            self.log(f"Translation error: {event.worker.error}")
            self.query_one("#output", Static).update("[red]Translation error[/]")
        elif event.state == WorkerState.CANCELLED:
            self.log("Translation cancelled")
    

def main():
    """Application entry point."""
    app = TranslationApp()
    app.run()


if __name__ == "__main__":
    main()



from .translation import (
    load_model,
    OnlineTranslation,
    TranslationModel,
    TimedText,
    MIN_SILENCE_DURATION_DEL_BUFFER,
)

from .core import TranslationBackend

from .languages import (
    get_nllb_code,
    get_language_code_code,
    get_language_name_by_language_code,
    get_language_name_by_nllb,
    get_language_info,
    list_all_languages,
    list_all_nllb_codes,
    list_all_language_code_codes,
    LANGUAGES,
)

__all__ = [
    # Main API
    "load_model",
    "OnlineTranslation",
    "TranslationModel",
    "TimedText",
    "MIN_SILENCE_DURATION_DEL_BUFFER",
    # Backend (advanced)
    "TranslationBackend",
    # Language utilities
    "get_nllb_code",
    "get_language_code_code",
    "get_language_name_by_language_code",
    "get_language_name_by_nllb",
    "get_language_info",
    "list_all_languages",
    "list_all_nllb_codes",
    "list_all_language_code_codes",
    "LANGUAGES",
]



import logging
import time
import torch
import transformers
from dataclasses import dataclass, field
from typing import Optional, Union
import huggingface_hub

try:
    import ctranslate2
    CTRANSLATE2_AVAILABLE = True
except ImportError:
    CTRANSLATE2_AVAILABLE = False
    ctranslate2 = None

from .languages import get_nllb_code, convert_to_nllb_code
from .core import TranslationBackend

logger = logging.getLogger(__name__)

MIN_SILENCE_DURATION_DEL_BUFFER = 1.0

@dataclass
class TimedText:
    start: Optional[float] = 0
    end: Optional[float] = 0
    text: Optional[str] = ''

    def overlaps_with(self, other: 'TimedText') -> bool:
        return not (self.end <= other.start or other.end <= self.start)

    def is_within(self, other: 'TimedText') -> bool:
        return other.contains_timespan(self)

    def approximate_cut_at(self, cut_time):
        """
        Each word in text is considered to be of duration (end-start)/len(words in text)
        """
        if not self.text or not self.contains_time(cut_time):
            return self, None

        words = self.text.split()
        num_words = len(words)
        if num_words == 0:
            return self, None

        duration_per_word = self.duration() / num_words
        
        cut_word_index = int((cut_time - self.start) / duration_per_word)
        
        if cut_word_index >= num_words:
            cut_word_index = num_words -1
        
        text0 = " ".join(words[:cut_word_index])
        text1 = " ".join(words[cut_word_index:])

        segment0 = Translation(start=self.start, end=cut_time, text=text0)
        segment1 = Translation(start=cut_time, end=self.end, text=text1)

        return segment0, segment1


@dataclass
class TranslationModel():
    translator: Union['ctranslate2.Translator', 'transformers.AutoModelForSeq2SeqLM']
    device: str
    tokenizer: dict = field(default_factory=dict)
    backend_type: str = 'transformers'
    nllb_size: str = '600M'
    model_name: str = field(default='')

    def get_tokenizer(self, input_lang):
        if not self.tokenizer.get(input_lang, False):
            model_name = self.model_name or f"facebook/nllb-200-distilled-{self.nllb_size}"
            self.tokenizer[input_lang] = transformers.AutoTokenizer.from_pretrained(
                model_name,
                src_lang=input_lang,
                clean_up_tokenization_spaces=True
            )
        return self.tokenizer[input_lang]


def load_model(src_langs, nllb_backend='transformers', nllb_size='600M'):
    device = "cuda" if torch.cuda.is_available() else "cpu"
    
    # Convert all source languages to NLLB codes
    converted_src_langs = []
    for lang in src_langs:
        if lang == 'auto':
            converted_src_langs.append('auto')
        else:
            nllb_code = convert_to_nllb_code(lang)
            if nllb_code is None:
                raise ValueError(f"Unknown language identifier: {lang}")
            converted_src_langs.append(nllb_code)

    if nllb_backend == 'ctranslate2': #BROKEN
        if not CTRANSLATE2_AVAILABLE:
            raise ImportError("ctranslate2 is not installed. Install it with: pip install ctranslate2")
        model = f'nllb-200-distilled-{nllb_size}-ctranslate2'
        MODEL_GUY = 'entai2965'
        huggingface_hub.snapshot_download(MODEL_GUY + '/' + model, local_dir=model)
        translator = ctranslate2.Translator(model, device=device)
        model_name = model
    elif nllb_backend == 'transformers':
        model_name = f"facebook/nllb-200-distilled-{nllb_size}"
        translator = transformers.AutoModelForSeq2SeqLM.from_pretrained(model_name).to(device)
    else:
        raise ValueError(f"Unknown backend: {nllb_backend}. Use 'transformers' or 'ctranslate2'")

    tokenizer = dict()
    for src_lang in converted_src_langs:
        if src_lang != 'auto':
            tokenizer[src_lang] = transformers.AutoTokenizer.from_pretrained(
                model_name if nllb_backend == 'transformers' else f"facebook/nllb-200-distilled-{nllb_size}",
                src_lang=src_lang,
                clean_up_tokenization_spaces=True
            )

    translation_model = TranslationModel(
        translator=translator,
        tokenizer=tokenizer,
        backend_type=nllb_backend,
        device=device,
        nllb_size=nllb_size,
        model_name=model_name if nllb_backend == 'transformers' else f"facebook/nllb-200-distilled-{nllb_size}"
    )
    return translation_model

class OnlineTranslation:
    def __init__(self, translation_model: TranslationModel, input_languages: list, output_languages: list):
        self.translation_model = translation_model        
        self.input_languages = []
        for lang in input_languages:
            if lang == 'auto':
                self.input_languages.append('auto')
            else:
                nllb_code = convert_to_nllb_code(lang)
                if nllb_code is None:
                    raise ValueError(f"Unknown input language identifier: {lang}")
                self.input_languages.append(nllb_code)
        
        self.output_languages = []
        for lang in output_languages:
            nllb_code = convert_to_nllb_code(lang)
            if nllb_code is None:
                raise ValueError(f"Unknown output language identifier: {lang}")
            self.output_languages.append(nllb_code)

        self.input_buffer = []
        self.last_buffer = ''
        self.commited = []
        self.len_computed: int = 0

        self.backend = TranslationBackend(
            source_lang=self.input_languages[0],
            target_lang=self.output_languages[0],
            model_name=translation_model.model_name,
            model=translation_model.translator,
            tokenizer=translation_model.get_tokenizer(self.input_languages[0])
        )

    def insert_tokens(self, tokens):
        self.input_buffer.extend(tokens)
    
    def process(self):
        text = ''.join([token.text for token in self.input_buffer])
        
        if self.input_buffer:
            start_time = self.input_buffer[0].start
            end_time = self.input_buffer[-1].end
        else:
            start_time = end_time = 0
        stable_translation, buffer_text = self.backend.translate(text)
        new_stable = stable_translation[self.len_computed:]
        self.len_computed += len(stable_translation)
        validated = TimedText(
            text=new_stable,
            start=start_time,
            end=end_time
        )

        buffer = TimedText(
            text=buffer_text,
            start=start_time,
            end=end_time
        )
        self.last_buffer = buffer
        self.commited.append(validated)
        return self.commited, buffer

    def insert_silence(self, silence_duration: float):
        if silence_duration >= MIN_SILENCE_DURATION_DEL_BUFFER:
            if self.last_buffer:
                if isinstance(self.last_buffer, str):
                    self.last_buffer = TimedText(text=self.last_buffer)
                self.commited.append(self.last_buffer)
            self.input_buffer = [] #maybe need to reprocess stuff before inserting silence
            self.last_buffer = ''



from translation_backend import TranslationBackend
import matplotlib.pyplot as plt
import pandas as pd

src_texts = [
    "Il s'arrêtait par moments devant une villa",
    "coquettement nichée dans la verdure, il regardait",
    "par la grille et voyait au loin des femmes",
    "élégantes sur les balcons et les terrasses, des",
    "enfants couraient dans les jardins. Il s'intéressait",
    "surtout aux fleurs ; c'étaient elles qui attiraient",
    "particulièrement ses regards. De temps en temps,",
    "il voyait passer des cavaliers, des amazones et de",
    "belles voitures ; il les suivait d'un œil curieux et",
    "les oubliait avant qu'ils eussent disparu. ",
    "Une fois, il s'arrêta et compta son argent ; il",
    "lui restait trente kopecks : « vingt au sergent de",
    "ville, trois à Nastassia pour la lettre, j'en ai donc",
    "donné hier à Marmeladov quarante-sept ou",
    "cinquante », se dit-il. Il devait avoir une raison de",
    "calculer ainsi, mais il l'oublia en tirant l'argent de",
    "sa poche et ne s'en souvint qu'un peu plus tard en",
    "passant devant un marchand de comestibles, une",
    "sorte de gargote plutôt ; il sentit alors qu'il avait",
    "faim. "
]

# print("1: direct generation)")
# translation_backend_1 = TranslationBackend(source_lang='fra_Latn', target_lang="eng_Latn")
# l_vals_no_cache = []

# for i in range(1, len(src_texts) + 1):
#     print(f'{i}/{len(src_texts) + 1}')
#     truncated_text = " ".join(src_texts[:i])
#     input_tokens = translation_backend_1.tokenizer(truncated_text, return_tensors="pt").to(translation_backend_1.device)
#     encoder_outputs = translation_backend_1.model.get_encoder()(**input_tokens)
#     output_tokens = translation_backend_1.model.generate(
#         encoder_outputs=encoder_outputs,
#         forced_bos_token_id=translation_backend_1.bos_token_id
#     )
#     output_text = translation_backend_1.tokenizer.decode(output_tokens[0], skip_special_tokens=True)
#     l_vals_no_cache.append({
#         "input": truncated_text,
#         "output_text": output_text,
#         "input_word_count": len(truncated_text.split()),
#         "output_word_count": len(output_text.split())
#     })

# print("2: prefix reuse")
# translation_backend_2 = TranslationBackend(source_lang='fra_Latn', target_lang="eng_Latn")
# l_vals_with_cache = []

# for i in range(1, len(src_texts) + 1):
#     print(f'{i}/{len(src_texts) + 1}')
#     truncated_text = " ".join(src_texts[:i])
#     stable_translation, buffer = translation_backend_2.translate(truncated_text)
    
#     full_output = stable_translation + buffer
#     l_vals_with_cache.append({
#         "input": truncated_text,
#         "stable_translation": stable_translation,
#         "buffer": buffer,
#         "full_output": full_output,
#         "input_word_count": len(truncated_text.split()),
#         "stable_word_count": len(stable_translation.split()) if stable_translation else 0,
#         "buffer_word_count": len(buffer.split()) if buffer else 0,
#         "total_output_word_count": len(full_output.split()) if full_output else 0
#     })

# df_base = pd.DataFrame(l_vals_no_cache)
# df_prefix = pd.DataFrame(l_vals_with_cache)
# df_base.to_csv('output_base_analysis.csv')
# df_prefix.to_csv('output_prefix_analysis.csv')

df_base = pd.read_csv('output_base_analysis.csv')
df_prefix = pd.read_csv('output_prefix_analysis.csv')

# df_merged = pd.concat([df_base, df_prefix], axis=1)

# Extract data from dataframe
iterations = list(range(1, len(df_prefix) + 1))
input_counts = df_prefix['input_word_count'].tolist()
output_with_prefix_and_buffer = df_prefix['total_output_word_count'].tolist()
stable_counts = df_prefix['stable_word_count'].tolist()
buffer_counts = df_prefix['buffer_word_count'].tolist()
base_output_wc = df_base['output_word_count'].tolist()

fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(14, 10))
fig, (ax1) = plt.subplots(1, 1, figsize=(5, 5))


colors = {
    "dark_gray_1": "#2E2E2E",   # deep graphite gray
    "dark_gray_2": "#4B4B4B",   # medium-dark neutral gray
    "orange": "#FF7F0E",        # modern vivid orange (Matplotlib's default orange)
    "red": "#D62728",           # elegant deep red
}
from matplotlib.ticker import MaxNLocator


ax1.plot(iterations, input_counts, label='Input', marker='o', color=colors['dark_gray_1'])
ax1.plot(iterations, base_output_wc, label='Base output (no prefix, no stability)', marker='^', color=colors['dark_gray_2'])
ax1.plot(iterations, stable_counts, label='With prefix (Stable part)', marker='^', linewidth=2, color=colors['red'])
ax1.plot(iterations, output_with_prefix_and_buffer, label='With prefix use output (Stable part + Buffer)', marker='^', color=colors['orange'])
ax1.set_xlabel('Iteration')
ax1.set_ylabel('Word count')
ax1.set_title('SimulMT french to english Word Count')
ax1.legend(fontsize=10)
ax1.grid(True, alpha=0.3)
ax1.xaxis.set_major_locator(MaxNLocator(integer=True))

# # Second plot: Detailed view of cached approach
# ax2.plot(iterations, input_counts, label='Input', marker='o', linewidth=2)
# ax2.plot(iterations, stable_counts, label='Stable prefix (cached)', marker='D', linewidth=2)
# ax2.plot(iterations, buffer_counts, label='Buffer (new)', marker='x', linewidth=2)
# ax2.plot(iterations, output_with_cache, label='Total output', marker='^', linewidth=2, linestyle='--')
# ax2.set_xlabel('Iteration', fontsize=12)
# ax2.set_ylabel('Word count', fontsize=12)
# ax2.set_title('Detailed view: Stable prefix vs Buffer (with cache)', fontsize=14, fontweight='bold')
# ax2.legend(fontsize=10)
# ax2.grid(True, alpha=0.3)

plt.tight_layout()
plt.savefig('translation_comparison.png', dpi=300, bbox_inches='tight')
plt.show()

# df = pd.DataFrame(l_vals_with_cache)
# df.to_csv('output_prefix_analysis.csv')
# print("terminated")



LANGUAGES = [
    {"name": "Acehnese (Arabic script)", "nllb": "ace_Arab", "language_code": "ace_Arab"},
    {"name": "Acehnese (Latin script)", "nllb": "ace_Latn", "language_code": "ace_Latn"},
    {"name": "Mesopotamian Arabic", "nllb": "acm_Arab", "language_code": "acm_Arab"},
    {"name": "Ta'izzi-Adeni Arabic", "nllb": "acq_Arab", "language_code": "acq_Arab"},
    {"name": "Tunisian Arabic", "nllb": "aeb_Arab", "language_code": "aeb_Arab"},
    {"name": "Afrikaans", "nllb": "afr_Latn", "language_code": "af"},
    {"name": "South Levantine Arabic", "nllb": "ajp_Arab", "language_code": "ajp_Arab"},
    {"name": "Akan", "nllb": "aka_Latn", "language_code": "ak"},
    {"name": "Tosk Albanian", "nllb": "als_Latn", "language_code": "als"},
    {"name": "Amharic", "nllb": "amh_Ethi", "language_code": "am"},
    {"name": "North Levantine Arabic", "nllb": "apc_Arab", "language_code": "apc_Arab"},
    {"name": "Modern Standard Arabic", "nllb": "arb_Arab", "language_code": "ar"},
    {"name": "Modern Standard Arabic (Romanized)", "nllb": "arb_Latn", "language_code": "arb_Latn"},
    {"name": "Najdi Arabic", "nllb": "ars_Arab", "language_code": "ars_Arab"},
    {"name": "Moroccan Arabic", "nllb": "ary_Arab", "language_code": "ary_Arab"},
    {"name": "Egyptian Arabic", "nllb": "arz_Arab", "language_code": "arz_Arab"},
    {"name": "Assamese", "nllb": "asm_Beng", "language_code": "as"},
    {"name": "Asturian", "nllb": "ast_Latn", "language_code": "ast"},
    {"name": "Awadhi", "nllb": "awa_Deva", "language_code": "awa"},
    {"name": "Central Aymara", "nllb": "ayr_Latn", "language_code": "ay"},
    {"name": "South Azerbaijani", "nllb": "azb_Arab", "language_code": "azb"},
    {"name": "North Azerbaijani", "nllb": "azj_Latn", "language_code": "az"},
    {"name": "Bashkir", "nllb": "bak_Cyrl", "language_code": "ba"},
    {"name": "Bambara", "nllb": "bam_Latn", "language_code": "bm"},
    {"name": "Balinese", "nllb": "ban_Latn", "language_code": "ban"},
    {"name": "Belarusian", "nllb": "bel_Cyrl", "language_code": "be"},
    {"name": "Bemba", "nllb": "bem_Latn", "language_code": "bem"},
    {"name": "Bengali", "nllb": "ben_Beng", "language_code": "bn"},
    {"name": "Bhojpuri", "nllb": "bho_Deva", "language_code": "bho"},
    {"name": "Banjar (Arabic script)", "nllb": "bjn_Arab", "language_code": "bjn_Arab"},
    {"name": "Banjar (Latin script)", "nllb": "bjn_Latn", "language_code": "bjn_Latn"},
    {"name": "Standard Tibetan", "nllb": "bod_Tibt", "language_code": "bo"},
    {"name": "Bosnian", "nllb": "bos_Latn", "language_code": "bs"},
    {"name": "Buginese", "nllb": "bug_Latn", "language_code": "bug"},
    {"name": "Bulgarian", "nllb": "bul_Cyrl", "language_code": "bg"},
    {"name": "Catalan", "nllb": "cat_Latn", "language_code": "ca"},
    {"name": "Cebuano", "nllb": "ceb_Latn", "language_code": "ceb"},
    {"name": "Czech", "nllb": "ces_Latn", "language_code": "cs"},
    {"name": "Chokwe", "nllb": "cjk_Latn", "language_code": "cjk"},
    {"name": "Central Kurdish", "nllb": "ckb_Arab", "language_code": "ckb"},
    {"name": "Crimean Tatar", "nllb": "crh_Latn", "language_code": "crh"},
    {"name": "Welsh", "nllb": "cym_Latn", "language_code": "cy"},
    {"name": "Danish", "nllb": "dan_Latn", "language_code": "da"},
    {"name": "German", "nllb": "deu_Latn", "language_code": "de"},
    {"name": "Southwestern Dinka", "nllb": "dik_Latn", "language_code": "dik"},
    {"name": "Dyula", "nllb": "dyu_Latn", "language_code": "dyu"},
    {"name": "Dzongkha", "nllb": "dzo_Tibt", "language_code": "dz"},
    {"name": "Greek", "nllb": "ell_Grek", "language_code": "el"},
    {"name": "English", "nllb": "eng_Latn", "language_code": "en"},
    {"name": "Esperanto", "nllb": "epo_Latn", "language_code": "eo"},
    {"name": "Estonian", "nllb": "est_Latn", "language_code": "et"},
    {"name": "Basque", "nllb": "eus_Latn", "language_code": "eu"},
    {"name": "Ewe", "nllb": "ewe_Latn", "language_code": "ee"},
    {"name": "Faroese", "nllb": "fao_Latn", "language_code": "fo"},
    {"name": "Fijian", "nllb": "fij_Latn", "language_code": "fj"},
    {"name": "Finnish", "nllb": "fin_Latn", "language_code": "fi"},
    {"name": "Fon", "nllb": "fon_Latn", "language_code": "fon"},
    {"name": "French", "nllb": "fra_Latn", "language_code": "fr"},
    {"name": "Friulian", "nllb": "fur_Latn", "language_code": "fur-IT"},
    {"name": "Nigerian Fulfulde", "nllb": "fuv_Latn", "language_code": "fuv"},
    {"name": "West Central Oromo", "nllb": "gaz_Latn", "language_code": "om"},
    {"name": "Scottish Gaelic", "nllb": "gla_Latn", "language_code": "gd"},
    {"name": "Irish", "nllb": "gle_Latn", "language_code": "ga-IE"},
    {"name": "Galician", "nllb": "glg_Latn", "language_code": "gl"},
    {"name": "Guarani", "nllb": "grn_Latn", "language_code": "gn"},
    {"name": "Gujarati", "nllb": "guj_Gujr", "language_code": "gu-IN"},
    {"name": "Haitian Creole", "nllb": "hat_Latn", "language_code": "ht"},
    {"name": "Hausa", "nllb": "hau_Latn", "language_code": "ha"},
    {"name": "Hebrew", "nllb": "heb_Hebr", "language_code": "he"},
    {"name": "Hindi", "nllb": "hin_Deva", "language_code": "hi"},
    {"name": "Chhattisgarhi", "nllb": "hne_Deva", "language_code": "hne"},
    {"name": "Croatian", "nllb": "hrv_Latn", "language_code": "hr"},
    {"name": "Hungarian", "nllb": "hun_Latn", "language_code": "hu"},
    {"name": "Armenian", "nllb": "hye_Armn", "language_code": "hy-AM"},
    {"name": "Igbo", "nllb": "ibo_Latn", "language_code": "ig"},
    {"name": "Ilocano", "nllb": "ilo_Latn", "language_code": "ilo"},
    {"name": "Indonesian", "nllb": "ind_Latn", "language_code": "id"},
    {"name": "Icelandic", "nllb": "isl_Latn", "language_code": "is"},
    {"name": "Italian", "nllb": "ita_Latn", "language_code": "it"},
    {"name": "Javanese", "nllb": "jav_Latn", "language_code": "jv"},
    {"name": "Japanese", "nllb": "jpn_Jpan", "language_code": "ja"},
    {"name": "Kabyle", "nllb": "kab_Latn", "language_code": "kab"},
    {"name": "Jingpho", "nllb": "kac_Latn", "language_code": "kac"},
    {"name": "Kamba", "nllb": "kam_Latn", "language_code": "kam"},
    {"name": "Kannada", "nllb": "kan_Knda", "language_code": "kn"},
    {"name": "Kashmiri (Arabic script)", "nllb": "kas_Arab", "language_code": "kas_Arab"},
    {"name": "Kashmiri (Devanagari script)", "nllb": "kas_Deva", "language_code": "kas_Deva"},
    {"name": "Georgian", "nllb": "kat_Geor", "language_code": "ka"},
    {"name": "Kazakh", "nllb": "kaz_Cyrl", "language_code": "kk"},
    {"name": "Kabiyè", "nllb": "kbp_Latn", "language_code": "kbp"},
    {"name": "Kabuverdianu", "nllb": "kea_Latn", "language_code": "kea"},
    {"name": "Halh Mongolian", "nllb": "khk_Cyrl", "language_code": "mn"},
    {"name": "Khmer", "nllb": "khm_Khmr", "language_code": "km"},
    {"name": "Kikuyu", "nllb": "kik_Latn", "language_code": "ki"},
    {"name": "Kinyarwanda", "nllb": "kin_Latn", "language_code": "rw"},
    {"name": "Kyrgyz", "nllb": "kir_Cyrl", "language_code": "ky"},
    {"name": "Kimbundu", "nllb": "kmb_Latn", "language_code": "kmb"},
    {"name": "Northern Kurdish", "nllb": "kmr_Latn", "language_code": "kmr"},
    {"name": "Central Kanuri (Arabic script)", "nllb": "knc_Arab", "language_code": "knc_Arab"},
    {"name": "Central Kanuri (Latin script)", "nllb": "knc_Latn", "language_code": "knc_Latn"},
    {"name": "Kikongo", "nllb": "kon_Latn", "language_code": "kg"},
    {"name": "Korean", "nllb": "kor_Hang", "language_code": "ko"},
    {"name": "Lao", "nllb": "lao_Laoo", "language_code": "lo"},
    {"name": "Ligurian", "nllb": "lij_Latn", "language_code": "lij"},
    {"name": "Limburgish", "nllb": "lim_Latn", "language_code": "li"},
    {"name": "Lingala", "nllb": "lin_Latn", "language_code": "ln"},
    {"name": "Lithuanian", "nllb": "lit_Latn", "language_code": "lt"},
    {"name": "Lombard", "nllb": "lmo_Latn", "language_code": "lmo"},
    {"name": "Latgalian", "nllb": "ltg_Latn", "language_code": "ltg"},
    {"name": "Luxembourgish", "nllb": "ltz_Latn", "language_code": "lb"},
    {"name": "Luba-Kasai", "nllb": "lua_Latn", "language_code": "lua"},
    {"name": "Ganda", "nllb": "lug_Latn", "language_code": "lg"},
    {"name": "Luo", "nllb": "luo_Latn", "language_code": "luo"},
    {"name": "Mizo", "nllb": "lus_Latn", "language_code": "lus"},
    {"name": "Standard Latvian", "nllb": "lvs_Latn", "language_code": "lv"},
    {"name": "Magahi", "nllb": "mag_Deva", "language_code": "mag"},
    {"name": "Maithili", "nllb": "mai_Deva", "language_code": "mai"},
    {"name": "Malayalam", "nllb": "mal_Mlym", "language_code": "ml-IN"},
    {"name": "Marathi", "nllb": "mar_Deva", "language_code": "mr"},
    {"name": "Minangkabau (Arabic script)", "nllb": "min_Arab", "language_code": "min_Arab"},
    {"name": "Minangkabau (Latin script)", "nllb": "min_Latn", "language_code": "min_Latn"},
    {"name": "Macedonian", "nllb": "mkd_Cyrl", "language_code": "mk"},
    {"name": "Maltese", "nllb": "mlt_Latn", "language_code": "mt"},
    {"name": "Meitei (Bengali script)", "nllb": "mni_Beng", "language_code": "mni"},
    {"name": "Mossi", "nllb": "mos_Latn", "language_code": "mos"},
    {"name": "Maori", "nllb": "mri_Latn", "language_code": "mi"},
    {"name": "Burmese", "nllb": "mya_Mymr", "language_code": "my"},
    {"name": "Dutch", "nllb": "nld_Latn", "language_code": "nl"},
    {"name": "Norwegian Nynorsk", "nllb": "nno_Latn", "language_code": "nn-NO"},
    {"name": "Norwegian Bokmål", "nllb": "nob_Latn", "language_code": "nb"},
    {"name": "Nepali", "nllb": "npi_Deva", "language_code": "ne-NP"},
    {"name": "Northern Sotho", "nllb": "nso_Latn", "language_code": "nso"},
    {"name": "Nuer", "nllb": "nus_Latn", "language_code": "nus"},
    {"name": "Nyanja", "nllb": "nya_Latn", "language_code": "ny"},
    {"name": "Occitan", "nllb": "oci_Latn", "language_code": "oc"},
    {"name": "Odia", "nllb": "ory_Orya", "language_code": "or"},
    {"name": "Pangasinan", "nllb": "pag_Latn", "language_code": "pag"},
    {"name": "Eastern Panjabi", "nllb": "pan_Guru", "language_code": "pa"},
    {"name": "Papiamento", "nllb": "pap_Latn", "language_code": "pap"},
    {"name": "Southern Pashto", "nllb": "pbt_Arab", "language_code": "pbt"},
    {"name": "Western Persian", "nllb": "pes_Arab", "language_code": "fa"},
    {"name": "Plateau Malagasy", "nllb": "plt_Latn", "language_code": "mg"},
    {"name": "Polish", "nllb": "pol_Latn", "language_code": "pl"},
    {"name": "Portuguese", "nllb": "por_Latn", "language_code": "pt-PT"},
    {"name": "Dari", "nllb": "prs_Arab", "language_code": "fa-AF"},
    {"name": "Ayacucho Quechua", "nllb": "quy_Latn", "language_code": "qu"},
    {"name": "Romanian", "nllb": "ron_Latn", "language_code": "ro"},
    {"name": "Rundi", "nllb": "run_Latn", "language_code": "rn"},
    {"name": "Russian", "nllb": "rus_Cyrl", "language_code": "ru"},
    {"name": "Sango", "nllb": "sag_Latn", "language_code": "sg"},
    {"name": "Sanskrit", "nllb": "san_Deva", "language_code": "sa"},
    {"name": "Santali", "nllb": "sat_Olck", "language_code": "sat"},
    {"name": "Sicilian", "nllb": "scn_Latn", "language_code": "scn"},
    {"name": "Shan", "nllb": "shn_Mymr", "language_code": "shn"},
    {"name": "Sinhala", "nllb": "sin_Sinh", "language_code": "si-LK"},
    {"name": "Slovak", "nllb": "slk_Latn", "language_code": "sk"},
    {"name": "Slovenian", "nllb": "slv_Latn", "language_code": "sl"},
    {"name": "Samoan", "nllb": "smo_Latn", "language_code": "sm"},
    {"name": "Shona", "nllb": "sna_Latn", "language_code": "sn"},
    {"name": "Sindhi", "nllb": "snd_Arab", "language_code": "sd"},
    {"name": "Somali", "nllb": "som_Latn", "language_code": "so"},
    {"name": "Southern Sotho", "nllb": "sot_Latn", "language_code": "st"},
    {"name": "Spanish", "nllb": "spa_Latn", "language_code": "es-ES"},
    {"name": "Sardinian", "nllb": "srd_Latn", "language_code": "sc"},
    {"name": "Serbian", "nllb": "srp_Cyrl", "language_code": "sr"},
    {"name": "Swati", "nllb": "ssw_Latn", "language_code": "ss"},
    {"name": "Sundanese", "nllb": "sun_Latn", "language_code": "su"},
    {"name": "Swedish", "nllb": "swe_Latn", "language_code": "sv-SE"},
    {"name": "Swahili", "nllb": "swh_Latn", "language_code": "sw"},
    {"name": "Silesian", "nllb": "szl_Latn", "language_code": "szl"},
    {"name": "Tamil", "nllb": "tam_Taml", "language_code": "ta"},
    {"name": "Tamasheq (Latin script)", "nllb": "taq_Latn", "language_code": "taq_Latn"},
    {"name": "Tamasheq (Tifinagh script)", "nllb": "taq_Tfng", "language_code": "taq_Tfng"},
    {"name": "Tatar", "nllb": "tat_Cyrl", "language_code": "tt-RU"},
    {"name": "Telugu", "nllb": "tel_Telu", "language_code": "te"},
    {"name": "Tajik", "nllb": "tgk_Cyrl", "language_code": "tg"},
    {"name": "Tagalog", "nllb": "tgl_Latn", "language_code": "tl"},
    {"name": "Thai", "nllb": "tha_Thai", "language_code": "th"},
    {"name": "Tigrinya", "nllb": "tir_Ethi", "language_code": "ti"},
    {"name": "Tok Pisin", "nllb": "tpi_Latn", "language_code": "tpi"},
    {"name": "Tswana", "nllb": "tsn_Latn", "language_code": "tn"},
    {"name": "Tsonga", "nllb": "tso_Latn", "language_code": "ts"},
    {"name": "Turkmen", "nllb": "tuk_Latn", "language_code": "tk"},
    {"name": "Tumbuka", "nllb": "tum_Latn", "language_code": "tum"},
    {"name": "Turkish", "nllb": "tur_Latn", "language_code": "tr"},
    {"name": "Twi", "nllb": "twi_Latn", "language_code": "tw"},
    {"name": "Central Atlas Tamazight", "nllb": "tzm_Tfng", "language_code": "tzm"},
    {"name": "Uyghur", "nllb": "uig_Arab", "language_code": "ug"},
    {"name": "Ukrainian", "nllb": "ukr_Cyrl", "language_code": "uk"},
    {"name": "Umbundu", "nllb": "umb_Latn", "language_code": "umb"},
    {"name": "Urdu", "nllb": "urd_Arab", "language_code": "ur"},
    {"name": "Northern Uzbek", "nllb": "uzn_Latn", "language_code": "uz"},
    {"name": "Venetian", "nllb": "vec_Latn", "language_code": "vec"},
    {"name": "Vietnamese", "nllb": "vie_Latn", "language_code": "vi"},
    {"name": "Waray", "nllb": "war_Latn", "language_code": "war"},
    {"name": "Wolof", "nllb": "wol_Latn", "language_code": "wo"},
    {"name": "Xhosa", "nllb": "xho_Latn", "language_code": "xh"},
    {"name": "Eastern Yiddish", "nllb": "ydd_Hebr", "language_code": "yi"},
    {"name": "Yoruba", "nllb": "yor_Latn", "language_code": "yo"},
    {"name": "Yue Chinese", "nllb": "yue_Hant", "language_code": "yue"},
    {"name": "Chinese (Simplified)", "nllb": "zho_Hans", "language_code": "zh-CN"},
    {"name": "Chinese (Traditional)", "nllb": "zho_Hant", "language_code": "zh-TW"},
    {"name": "Standard Malay", "nllb": "zsm_Latn", "language_code": "ms"},
    {"name": "Zulu", "nllb": "zul_Latn", "language_code": "zu"},
]

NAME_TO_NLLB = {lang["name"]: lang["nllb"] for lang in LANGUAGES}
NAME_TO_LANGUAGE_CODE = {lang["name"]: lang["language_code"] for lang in LANGUAGES}
LANGUAGE_CODE_TO_NLLB = {lang["language_code"]: lang["nllb"] for lang in LANGUAGES}
NLLB_TO_LANGUAGE_CODE = {lang["nllb"]: lang["language_code"] for lang in LANGUAGES}
LANGUAGE_CODE_TO_NAME = {lang["language_code"]: lang["name"] for lang in LANGUAGES}
NLLB_TO_NAME = {lang["nllb"]: lang["name"] for lang in LANGUAGES}


def get_nllb_code(language_code_code):
    return LANGUAGE_CODE_TO_NLLB.get(language_code_code, None)


def get_language_code_code(nllb_code):
    return NLLB_TO_LANGUAGE_CODE.get(nllb_code)


def get_language_name_by_language_code(language_code_code):
    return LANGUAGE_CODE_TO_NAME.get(language_code_code)


def get_language_name_by_nllb(nllb_code):
    return NLLB_TO_NAME.get(nllb_code)


def get_language_info(identifier, identifier_type="auto"):
    if identifier_type == "auto":
        for lang in LANGUAGES:
            if (lang["name"].lower() == identifier.lower() or 
                lang["nllb"] == identifier or 
                lang["language_code"] == identifier):
                return lang
    elif identifier_type == "name":
        for lang in LANGUAGES:
            if lang["name"].lower() == identifier.lower():
                return lang
    elif identifier_type == "nllb":
        for lang in LANGUAGES:
            if lang["nllb"] == identifier:
                return lang
    elif identifier_type == "language_code":
        for lang in LANGUAGES:
            if lang["language_code"] == identifier:
                return lang
    
    return None


def convert_to_nllb_code(language_identifier):
    if language_identifier == 'auto':
        return 'auto'
    
    language_info = get_language_info(language_identifier)
    if language_info:
        return language_info["nllb"]
    
    return None


def list_all_languages():
    return [lang["name"] for lang in LANGUAGES]


def list_all_nllb_codes():
    return [lang["nllb"] for lang in LANGUAGES]


def list_all_language_code_codes():
    return [lang["language_code"] for lang in LANGUAGES]



import torch
from transformers import AutoTokenizer, AutoModelForSeq2SeqLM
from transformers.cache_utils import EncoderDecoderCache, DynamicCache
from typing import Tuple, Optional


class TranslationBackend:
    def __init__(self, source_lang, target_lang, model_name: str = "facebook/nllb-200-distilled-600M", model=None, tokenizer=None):
        self.device = "cuda" if torch.cuda.is_available() else "cpu"

        if model is not None:
            self.model = model
            if not hasattr(model, 'device') or str(model.device) != self.device:
                self.model = self.model.to(self.device)
        else:
            self.model = AutoModelForSeq2SeqLM.from_pretrained(model_name).to(self.device)

        if tokenizer is not None:
            self.tokenizer = tokenizer
        else:
            self.tokenizer = AutoTokenizer.from_pretrained(model_name, src_lang=source_lang)
        
        self.source_lang = source_lang
        self.target_lang = target_lang
        self.bos_token_id = self.tokenizer.convert_tokens_to_ids(self.target_lang)

        self.sentence_end_token_ids = set()

        # find all tokens that decode to sentence-ending punctuation. For ex: token 81 and 248075 both represent '.')
        for token_id in range(min(300000, self.tokenizer.vocab_size)):
            try:
                decoded = self.tokenizer.decode([token_id])
                cleaned = decoded.strip().strip("'\"").strip()
                if cleaned in ['.', '!', '?']:
                    self.sentence_end_token_ids.add(token_id)
            except:
                pass

        self.previous_tokens = None
        self.stable_prefix = None
    
    def simple_translation(self, text):
        inputs = self.tokenizer(text, return_tensors="pt").to(self.device)
        encoder_outputs = self.model.get_encoder()(**inputs)
        output = self.generate(
                    encoder_outputs=encoder_outputs,
        )
        result = self.tokenizer.decode(output[0], skip_special_tokens=True)
        return output, result

    def generate(
        self,
        encoder_outputs: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        max_length: Optional[int] = 200
        
    ) -> Tuple[torch.Tensor, Optional[Tuple]]:    
        with torch.no_grad():
            generated_tokens = self.model.generate(
                encoder_outputs=encoder_outputs,
                attention_mask=attention_mask,
                forced_bos_token_id=self.bos_token_id,
                max_length=max_length,
            )
        return generated_tokens

    def compute_common_prefix_tokens(
        self, new_tokens
    ):
        common_length = 0
        for i in range(min(len(self.previous_tokens[0]), len(new_tokens[0]))):
            if self.previous_tokens[0][i] != new_tokens[0][i]:
                common_length = i
                break
        else:
            common_length = min(len(self.previous_tokens[0]), len(new_tokens[0]))

        last_sentence_end = -1
        for i in range(common_length):
            if new_tokens[0][i].item() in self.sentence_end_token_ids:
                last_sentence_end = i

        if last_sentence_end >= 0:
            return new_tokens[:, :last_sentence_end]

        return new_tokens[:, :common_length]
    
    def translate(self, text: str) -> str:
        word_count = len(text.strip().split())
        
        if word_count < 3:
            return "", ""
    
        inputs = self.tokenizer(text, return_tensors="pt").to(self.device)            
        with torch.no_grad():
            encoder_outputs = self.model.get_encoder()(**inputs)
        
        if (self.previous_tokens is not None and self.stable_prefix is not None):
            translation_tokens = self._continue_generation_with_cache(
                encoder_hidden_states=encoder_outputs.last_hidden_state,
            )
        else:
            with torch.no_grad():
                translation_tokens = self.generate(
                    encoder_outputs=encoder_outputs,
                    attention_mask=inputs['attention_mask'],
                )
        if self.previous_tokens is not None:
            self.stable_prefix = self.compute_common_prefix_tokens(new_tokens=translation_tokens)
        self.previous_tokens = translation_tokens
        
        buffer = self.tokenizer.decode(
                translation_tokens[0], 
                skip_special_tokens=True
        )        
        if self.stable_prefix is not None:
            stable_translation = self.tokenizer.decode(
                self.stable_prefix[0], 
                skip_special_tokens=True
            )
            return stable_translation, buffer[len(stable_translation):]            
        else:
            return "", buffer

    def _continue_generation_with_cache(
        self,
        encoder_hidden_states: torch.Tensor,
        max_new_tokens: int = 200
    ) -> torch.Tensor:
        eos_token_id = self.tokenizer.eos_token_id

        with torch.no_grad():
            past_key_values = EncoderDecoderCache(
                self_attention_cache=DynamicCache(),
                cross_attention_cache=DynamicCache()
            )

            decoder_out = self.model.model.decoder(
                input_ids=self.stable_prefix,
                encoder_hidden_states=encoder_hidden_states,
                past_key_values=past_key_values,
                use_cache=True,
                return_dict=True,
            )
            past_key_values = decoder_out.past_key_values
            prefix_logits = self.model.lm_head(decoder_out.last_hidden_state)
            next_token_id = torch.argmax(prefix_logits[:, -1, :], dim=-1).unsqueeze(-1)

            generated_tokens = self.stable_prefix.clone()

            for _ in range(max_new_tokens):
                if next_token_id.item() == eos_token_id:
                    break

                generated_tokens = torch.cat([generated_tokens, next_token_id], dim=-1)

                decoder_out = self.model.model.decoder(
                    input_ids=next_token_id,
                    encoder_hidden_states=encoder_hidden_states,
                    past_key_values=past_key_values,
                    use_cache=True,
                    return_dict=True,
                )
                past_key_values = decoder_out.past_key_values
                logits = self.model.lm_head(decoder_out.last_hidden_state)
                next_token_id = torch.argmax(logits[:, -1, :], dim=-1).unsqueeze(-1)

        return generated_tokens



if __name__ == '__main__':
    # src_texts = [
    #     "Have you noticed how accurate",
    #     "LLM are now that GPU have became more powerful, ",
    #     "especially when we think at the difficulties we had in the 2010 era",
    #     "where online chatbots were not performant",
    #     "at all, and ofter doing strict rules worked better",
    #     "do you remember that era?"]

#     src_texts = [
#         "As-tu remarqué à quel point",
#         "les LLM sont précis maintenant que les GPU sont devenus plus puissants, ",
#         "surtout quand on pense aux difficultés qu'on avait dans les années 2010",
#         "où les chatbots en ligne n'étaient pas performants",
#         "du tout, et souvent faire des règles strictes fonctionnait mieux.",
#         "Tu te souviens de cette époque, ",
#         "c'était bien plus compliqué de travailler"]

    translation_backend = TranslationBackend(source_lang='fra_Latn', target_lang="eng_Latn")
    
#     # for i in range(len(src_texts)+1):
#     #     translation, buffer = translation_backend.translate(" ".join(src_texts[:i]))
#     #     print(translation, '|', buffer)

#     tokens, text = translation_backend.simple_translation("Ceci est un test de traduction. Nous avons fait tout notre possible. Jusqu'à partir de 0")[1]
    
#     # text = "Ceci est un test de traduction. Nous avons fait tout notre possible. Jusqu'à partir de 0, où nous avions pris la route, tout droit, et loin, loin, loin ! tu penses qu'on peut vraiment faire la longeur qu'on veut ?"
#     print(output_text) 
    

    src_texts = ["Il s’arrêtait par moments devant une villa",
    "coquettement nichée dans la verdure, il regardait",
    "par la grille et voyait au loin des femmes",
    "élégantes sur les balcons et les terrasses, des",
    "enfants couraient dans les jardins. Il s’intéressait",
    "surtout aux fleurs ; c’étaient elles qui attiraient",
    "particulièrement ses regards. De temps en temps,"
    "il voyait passer des cavaliers, des amazones et de"
    "belles voitures ; il les suivait d’un œil curieux et",
    "les oubliait avant qu’ils eussent disparu. "]

    text = " ".join(src_texts)
    

    target = """He would stop occasionally in front of a villa
    nestled charmingly in the greenery, look
    through the gate, and see elegant women
    on balconies and terraces in the distance,
    children running in the gardens. He was particularly interested
    in the flowers; they were what caught his eye. From time to time, 
    he would see horsemen, horsewomen, and beautiful carriages passing by; 
    he would follow them with a curious eye and
    forget them before they had disappeared.
    """
    l_vals = []
    for i in range(1, len(src_texts)+1):
        truncated_text = " ".join(src_texts[:i])
        input_tokens = translation_backend.tokenizer(truncated_text, return_tensors="pt").to(translation_backend.device)
        encoder_outputs = translation_backend.model.get_encoder()(**input_tokens)
        output_tokens = translation_backend.model.generate(
            encoder_outputs=encoder_outputs,
            forced_bos_token_id=translation_backend.bos_token_id
        )
        output_text = translation_backend.tokenizer.decode(output_tokens[0], skip_special_tokens=True)
        l_vals.append(
            {
                "input": truncated_text,
                "output_tokens_shape": output_tokens.shape[1], 
                "output_text": output_text,
            }
        )
    # input_tokens = translation_backend.tokenizer(text, return_tensors="pt").to(translation_backend.device)
    # encoder_outputs = translation_backend.model.get_encoder()(**input_tokens)
    # output_tokens = translation_backend.model.generate(
    #     encoder_outputs=encoder_outputs,
    #     forced_bos_token_id=translation_backend.bos_token_id
    # )
    # output_text = translation_backend.tokenizer.decode(output_tokens[0], skip_special_tokens=True)

    
    print('end')
